/** \file generate.c
*/
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>

#include "generate.h"
#include "nv_hw.h"
#include "objects.h"
#include "properties.h"
/** \mainpage
   
   Currently the default resource file is not read in. Please use
   the cli options instead:
     -cv  comments verbose
     -cs  short comments [default]
     -cn  no comments except for objects use on <Card>
     -a <Name>  The author's name / Email address
     -gv  add "generated from version" to header [default]
     -gnv don't add generated from version to header
     -t   add timestamp to header
     -i   include "include.txt" into the header just beneath the #ifndef 
*/ 
//#include "objects.c"

/*

Some recent updates requires this. But as it is only needed for printing results
in renouveau while we are only interested in the structure nv_objects[] we hack
around this (It is defined in main.c which we can't import / link here)

*/ 
#define MAXPRINT 40
/* Used by object.c */
data_print_store user_data_print[MAXPRINT];


int fifo_channel = -1;
extern const object_store nv_objects[];
t_properties *props;
static int current_column;
nv_exclude_objects exclude_objects[] = 
      { /*{"NV10_TCL_PRIMITIVE_3D"}, 
        {"NV17_TCL_PRIMITIVE_3D"}, */
        {"\0"} };

/**
   \fn    void str2upper(char *buffer)
   \brief convert lowercase fieldnames to uppercase and change spaces into "_"
*/
void str2upper(char *buffer)
{
  char *ptr;

  ptr = buffer;
  while (*ptr != (char) 0)
   {
     *ptr = toupper(*ptr);
     if (*ptr == ' ')
       {
        *ptr = '_';
       }
	 /* define is alphanum and _ , transform everything to _*/
     else if (isalnum(*ptr) == 0)
       {
        *ptr = '_';
       }

     ptr++;
   }
}

void get_time(char *buffer)
{
}
  
void create_header(int header_file)
{
  int    filed, count;
  char   buffer[2000], version[30], tbuff[40];
  time_t cr_time;
  
  filed = open(HEADER_LICENSE_FILENAME, O_RDONLY);
  if (filed != -1)
     {
       do 
       {
         write_text(header_file, "/*\n");
         count = read(filed, buffer, 2000);
         write(header_file, buffer, count);
         if (props->pr_version == 1)
           {
             get_version_line(version);
             snprintf(buffer, 2000, "   Created from objects.c rev. %s\n", version);
             write_text(header_file, buffer);
           }
       } while (count == 2000);
       write_text(header_file, "*/\n");
       close(filed);
     }
    else 
     {
       fprintf(stderr, "WARNING: No license.txt found. Creating short version\n");
       write_text(header_file, "/* nouveau_reg.h\n");
       write_text(header_file, "   Autogenerated file from objects.c and nv_objects.h\n");
       write_text(header_file, "   don't touch or risk to loose your changes\n");
       write_text(header_file, "   (C) 2006 Nouveau team\n");
       if (props->pr_version == 1)
         {
           get_version_line(version);
           if (props->pr_time == 0)
              snprintf(buffer, 2000, "\n   Created from objects.c rev. %s\n", version);
            else 
             { 
               cr_time = time(NULL);
               ctime_r(&cr_time, tbuff);
               snprintf(buffer, 2000, "\n   Created from objects.c rev. %s at %s\n", version, tbuff);
             }
           write_text(header_file, buffer);
         }
       if (props->pr_author[0] != (char) 0)
          {
            write_text(header_file, "   Author: ");
            write_text(header_file, props->pr_author);
            write_text(header_file, "\n");
          }
       write_text(header_file, "*/\n");
         
     }
  write_text(header_file, "\n#ifndef _NOUVEAU_REG_H\n");
  write_text(header_file, "#define _NOUVEAU_REG_H\n");
}

void create_default_include(int filedw)
{
  char buffer[2000];
  int  count, filedr;
  
  if (props->pr_include[0] != (char) 0)
     {
         filedr = open(props->pr_include, O_RDONLY);
         if (filedr == -1)
            {
              fprintf(stderr, "ERROR: Can't open %s: %s\n", props->pr_include,
                      strerror(errno));
              return;
            }
         do
         {
             count = read(filedr, buffer, 2000);
             write(filedw, buffer, count);
             write_text(filedw, buffer);
         } while (count == 2000);
         write_text(filedw, "*/\n");
         close(filedr);
     }
}

void create_bottom(int header_file)
{
  write_text(header_file, "\n#endif /* _NOUVEAU_REG_H */\n");
  close(header_file);
}

void create_object_define(int filed, object_store obj)
{
   char buffer[100];
   write_text(filed, "#define\t\t\t");
   write_text(filed, obj.name);
   write_align(filed, 80);
   snprintf(buffer, 100, "0x%08x", obj.id);
   write_text(filed, buffer);
   write_text(filed, "\n");
}

void create_object_comment(int filed, object_store obj)
{
   
   write_text(filed, "\n/****************************************** \nObject ");
   write_text(filed, obj.name);
   write_text(filed, " used on: ");
   if ( obj.card_type & NV03)
       write_text(filed, "NV03 ");
   if ( obj.card_type & NV04)
       write_text(filed, "NV04 ");
   if ( obj.card_type & NV10)
       write_text(filed, "NV10 ");
   if ( obj.card_type & NV15)
       write_text(filed, "NV15 ");
   if ( obj.card_type & NV20)
       write_text(filed, "NV20 ");
   if ( obj.card_type & NV30)
       write_text(filed, "NV30 ");
   if ( obj.card_type & NV40)
       write_text(filed, "NV40 ");
   if ( obj.card_type & G70)
       write_text(filed, "G70");
   write_text(filed, "\n*/\n");
}

void write_text(int filed, const char *text)
{
  write (filed, text, strlen(text));
  int i;
  for(i=0;i<strlen(text);i++)
  {
    if (text[i]=='\t')
      current_column=(current_column+TAB_SPACE)-(current_column%TAB_SPACE);
    else if (text[i]=='\n')
      current_column=0;
    else
      current_column++;
  }
}

void write_align(int filed,int column)
{
  write_text(filed,"\t");
  while(current_column<column)
    write_text(filed,"\t");
}

int getArrayType(char *name)
{
  char buffer[10];
  int  i;
  printf("   %s\n", name);
  // Simple hack
  // 1. Let's look for a matrix
  if (strstr(name, "(0)(%d)") != NULL)
     return 4;
  if (strstr(name, "(%d)(0)") != NULL)
     return 4;
  for (i = 1; i < MAX_INDEX; i++)
   {
    sprintf(buffer, "(%d)(%%d)", i);
    if (strstr(name, buffer) != NULL)
       return 3;
    sprintf(buffer, "(%%d)(%d)", i);
    if (strstr(name, buffer) != NULL)
       return 3;
    sprintf(buffer, "(%d)", i);
    if (strstr(name, buffer) != NULL)
       return 3;
   }
  // Ok, either manual Array or A_() array
  if (strstr(name, "(%d)") != NULL)
     return 1;
  if (strstr(name, "(0)") != NULL)
     return 2;
   return 0;
}

void create_line_comment(int filed, object_field_store field)
{
  int i;
  
  if (props->pr_comments != COMMENTS_SHORT)
     return;
  if (field.data[0].name == NULL)
     return;
  write_text(filed, "\t/* Parameters:");
  for (i = 0; field.data[i].name != NULL; i++)
    { 
      write_text(filed, " ");
      write_text(filed, field.data[i].name);
    }
  write_text(filed, " */");
}

void create_object_commands(int filed, object_store obj)
{
  int   i, offset2, crwrite, offsetm;
  char  buffer[100], *ptr;

  for (i = 0; obj.fields[i].name != NULL; i++)
     {
       crwrite = 1;
       switch(getArrayType((char *)obj.fields[i].name))
         {
           case 0: // Normal define
                 create_command_comment(filed, obj.fields[i]);
                 write_text(filed, "#");
                 write_text(filed, INCLUDE_IDENT_LVL1);
                 write_text(filed, "define\t\t");
                 write_text(filed, obj.fields[i].name);
                 write_align(filed, COLUMN_ALIGN);
                 snprintf(buffer,100, "0x%08x", obj.fields[i].offset);
                 write_text(filed, buffer);
                 break;
           case 1: // A_() array. Simple to define
                 create_command_comment(filed, obj.fields[i]);
                 write_text(filed, "#");
                 write_text(filed, INCLUDE_IDENT_LVL1);
                 write_text(filed, "define\t\t");
                 strcpy(buffer, obj.fields[i].name);
                 ptr = strchr(buffer, '%'); // Can't be NULL
                 *ptr = ' ';
                 write_text(filed, buffer);
                 write_align(filed, COLUMN_ALIGN);
                 snprintf(buffer, 100, "(0x%08x + (d) * 0x%04x)", 
                          obj.fields[i].offset, obj.fields[i].stride);
                 write_text(filed, buffer);
                 break;
           case 2:  // Basic array manually defined
                 create_command_comment(filed, obj.fields[i]);
                 write_text(filed, "#");
                 write_text(filed, INCLUDE_IDENT_LVL1);
                 write_text(filed, "define\t\t");
                 // 1. get second offset
                 offset2 = get_index_offset((char *) obj.fields[i].name, obj);
                 // 2. replace (0) with (d)
                 strcpy(buffer, obj.fields[i].name);
                 ptr = strstr(buffer, "(0)");
                 ptr++;
                 *ptr = 'd';
                 write_text(filed, buffer);
                 write_align(filed, COLUMN_ALIGN);
                 // 3. get stride by offset(1) - offset(0)
                 snprintf(buffer, 100, "(0x%08x + (d) * 0x%04x)", 
                          obj.fields[i].offset, 
                          offset2 - obj.fields[i].offset);
                 // 4. write data name(d) = (offset(0) + d * stride)
                 write_text(filed, buffer);
                 break;
           case 3:  // Manually defined, index != 0
                 crwrite = 0;
                 break;
           case 4:  // A matrix
                 write_text(filed, "#");
                 write_text(filed, INCLUDE_IDENT_LVL1);
                 write_text(filed, "define\t\t");
                 create_command_comment(filed, obj.fields[i]);
                 offsetm = get_matrix_offset((char *) obj.fields[i].name, obj);
                 offset2 = get_index_offset((char *) obj.fields[i].name, obj);
                 strcpy(buffer, obj.fields[i].name);
                 ptr = strchr(buffer, '(');
                 *ptr = (char) 0;
                 strcat(buffer, "(x,y)");
                 write_text(filed, buffer);
                 write_align(filed, COLUMN_ALIGN);
                 snprintf(buffer, 100, "(0x%08x + (y) * 0x0010 + (x) * 0x%04x)",
                          obj.fields[i].offset, 
                          obj.fields[i].stride);
                 write_text(filed, buffer);
                 break;
         }
       if (crwrite)
          {
            if (props->pr_comments != COMMENTS_NONE)
               create_line_comment(filed, obj.fields[i]);
            write_text(filed, "\n");
            create_masks(filed, obj.fields[i]);

          }
     }
}

void dump_decimal_mask(int filed, data_print_store store)
{
   int  count = 0;
   int  i;
   char format[30];
   
   if ((store.lowbit != 0) ||
       (store.highbit != 31))
      { 
        for (i = store.lowbit; i <= store.highbit; i++)
          count += (1 << i);
        snprintf(format, 30, "mask 0x%08x", count);
        write_text(filed, format);
      }

}

void create_command_parameter_comment(int filed, object_field_store field)
{
  int  i,j;

  write_text(filed, "\nParameters used:\n");
  for (i = 0; field.data[i].name != NULL; i++)
    { 
      write_text(filed, "  ");
      write_text(filed, field.data[i].name);
      write_text(filed, " : ");
      // lowbit, highbit
      if (field.data[i].func == print_float)
         write_text(filed, "FP32");
        else
      if (field.data[i].func == print_format)
         {
           write_text(filed, "Decimal ");
           dump_decimal_mask(filed, field.data[i]);
         }
        else
      if (field.data[i].func == print_enable)
        {
          write_text(filed, "Boolean");
        }
       else
      if (field.data[i].func == print_macro)
        {
         data_store *names = (data_store *)field.data[i].data;
          if (names != NULL)
             {
//               write_text(filed, " names exist");
             }
	  j = 0;

   	  while (names[j].name != NULL) 
   	    {
		write_text(filed, names[j].name);
		write_text(filed, " ");
		j++;
  	    }
        }
        
      write_text(filed, "\n");
    }
  
}

void create_command_comment(int filed, object_field_store field)
{
 
  if (props->pr_comments != COMMENTS_VERBOSE)
     return;  
  if ((field.data[0].name == NULL)) // &&
//      ((field.size == 1) && (field.stride ==1)))
      return;
  write_text(filed, "/*\n");
/*  if ((field.size != 1) || (field.stride != 1))
     {
      snprintf(buffer,  200,  "Size is %d, stride is %d\n", field.size, field.stride);
      write_text(filed, buffer);
     }
*/     
  if (field.data[0].name != NULL)
    {
     create_command_parameter_comment(filed, field);
    }
  write_text(filed, "*/\n");
}

int get_index_offset(char * name, object_store obj)
{
  int   i;
  char *ptr, buffer[100];
  
  strcpy(buffer, name);
  ptr = strchr(buffer, '(');
  ptr ++;
  *ptr = '1';
  for (i = 0; obj.fields[i].name != NULL; i++)
    {
      if (strstr(obj.fields[i].name, buffer) != NULL)
          return obj.fields[i].offset;
    }
  fprintf(stderr, "ERROR: Array %s not found\n", name);
  return -1;
}

int get_matrix_offset(char * name, object_store obj)
{
  int   i;
  char *ptr, buffer[100];
  
  strcpy(buffer, name);
  ptr = strchr(buffer, '(');
  ptr ++;
  *ptr = '1';
  for (i = 0; obj.fields[i].name != NULL; i++)
    {
      if (strstr(obj.fields[i].name, buffer) != NULL)
          return obj.fields[i].offset;
    }
  fprintf(stderr, "ERROR: Matrix %s not found\n", name);
  return -1;
}

int check_exclusion(char *obj_name)
{
  int i;
  
  i = 0;
  while (exclude_objects[i].ex_name[0] != (char) 0)
   {
     if (strcmp(exclude_objects[i].ex_name, obj_name) == 0)
       {
        printf("Skipping duplicate %s\n", exclude_objects[i].ex_name); 
        return 1;
       }
     i++;
   }
  return 0;
}


int main(int argc, char *argv[])
{
  int  header_file,i ;
  
  props = getProperties(argc, argv);
  printf("Creating %s please wait\n", HEADER_FILENAME); 
  header_file = open(HEADER_FILENAME, O_WRONLY | O_TRUNC | O_CREAT,
                     S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
  if (header_file == -1)
     {
       fprintf(stderr, "Cannot create %s\n", HEADER_FILENAME);
       fprintf(stderr, strerror(errno));
       exit(1);
     }
  create_header(header_file);
  create_default_include(header_file);
  for (i = 0; nv_objects[i].name != NULL; i++)
    {
      create_object_comment(header_file, nv_objects[i]);
      create_object_define(header_file, nv_objects[i]);
      if (check_exclusion((char *) nv_objects[i].name))
         continue;
/*      if (strcmp(nv_objects[i].name,"NV10_TCL_PRIMITIVE_3D")==0) {
        continue;
      }
      if (strcmp(nv_objects[i].name,"NV17_TCL_PRIMITIVE_3D")==0) {
        continue;
      }
*/
      create_object_commands(header_file, nv_objects[i]);
      fprintf(stdout, "Creating %s...\n", nv_objects[i].name);
    }
  create_bottom(header_file);
  return 0;
}

void strip_array(char *fieldname, char *result)
{
  char *ptr;

  strcpy(result, fieldname);
  ptr = strstr(result, "(0)");
  if (ptr != NULL)
     *ptr = (char) 0;
}

void create_masks(int filed, object_field_store field)
{
  int  i;
  char buff[100], dataname[100], fname[100], *ptr;
  
  if (field.data[0].name == NULL)
     return;
  if (props->pr_comments != COMMENTS_SHORT)
  {
  	write_text(filed, "\n/* MASKS:");
  	// First for only for debugging, if you feel like it : remove it ;)
  	for (i = 0; field.data[i].name != NULL; i++)
	{
		write_text(filed, " ");
		write_text(filed, field.data[i].name);
		snprintf(buff, 100, " Low %d, High %d ", field.data[i].lowbit,field.data[i].highbit);
		write_text(filed, buff);
		snprintf(buff, 100, " Mask 0x%08x\n", get_mask(field.data[i].lowbit,field.data[i].highbit));
		write_text(filed, buff);
	}
  	write_text(filed, "\n");
  	write_text(filed, " */\n");
  }
  //
  // Now the data we really want to see:
  //
  for (i = 0; field.data[i].name != NULL; i++)
    {
      strip_array(field.name, fname);
      write_text(filed, "#\t\t define\t\t");
      strcpy(dataname, field.data[i].name);
      str2upper(dataname);
      ptr = strchr(fname, '(');
      if (ptr != NULL)
      {
	*ptr = (char) 0;
      }
      snprintf(buff, 100, "%s_%s_MASK\t0x%08x\n", fname, dataname,
               get_mask(field.data[i].lowbit,field.data[i].highbit));
      write_text(filed, buff);
      // Do we need _SHIFT?
      if (field.data[i].lowbit == 0)
         continue;  // NO
      // Check whether field length is longer than 1 Bit
      write_text(filed, "#\t\t define\t\t");
      strcpy(dataname, field.data[i].name);
      str2upper(dataname);
      if (field.data[i].lowbit != field.data[i].highbit)
        {
         snprintf(buff, 100, "%s_%s_SHIFT\t%d\n", fname, dataname, field.data[i].lowbit);
        }
       else
        {
         if (field.data[i].lowbit == 1)
            snprintf(buff, 100, "%s_%s\t1 // Nothing to shift\n", fname, dataname);
          else snprintf(buff, 100, "%s_%s\t(1 << %d)\n", fname, dataname, field.data[i].lowbit);
        }
      write_text(filed, buff);
      // Do we have data store?
      if (field.data[i].data != NULL)
         {  // Maybe, but only if we don't print_format (in this case data is a string)
           if (field.data[i].func != print_format)
             {
	           data_store *ptr;
        	   ptr	 = field.data[i].data;
                   while (ptr->name != NULL)
                    {
     		      snprintf(buff,100, "#\t\t define %s_%s_%s\t0x%04x\n", fname, dataname, ptr->name, ptr->id);
                      write_text(filed, buff);
                      ptr++;
                    }
             }
         }
    }

}

int get_mask(int low, int high)
{
  int result, i;

  result = 0;
  for (i = low; i <= high; i++)
   {
     result = result | (1 << i);
   }
  return result;
}

